/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#ifndef lint
static const char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 04\n\
@(#)All rights reserved.\n";
static const char rcsid[] = "$Id: pgoblin.c,v 1.40 2008/05/31 10:16:33 babolo Exp $";
#endif /* not lint */

#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <mife.h>
#include <err.h>
#include <babolo/parser.h>
#include <babolo/BLINflag.h>
#include "pgoblin.h"

static void
usage(int ex) {
    int i;
    char *q;

    fprintf(stderr, "pgoblin " VERS "  " DATE "\n"
"Usage:\n"
"pgoblin [-aANOqvV ][-d DataBase ][-h Hostname ][-p Port ][-U User ][-f QueryFile][-Eflag[env]] args\n"
"    -f QueryFile\n"
"    -d DataBase name\n"
"    -h Hostname of backend server\n"
"    -p Port\n"
"____-U Username\n"
"    -c COPY out\n"
"    -C COPY in\n"
"    -D perform query without result\n"
"    -s SELECT\n"
"    -S SELECT BINARY emulation\n"
"____-e QueryArgument\n"
"    -q quiet\n"
"    -v verbose\n"
"    -N new style\n"
"    -O old style\n"
"Hooked up:\n"
);
    for (i = 0; DBASE(i); i++) {
        q = index(DBASE(i)->name, '\0');
        fprintf(stderr, "  %s %s %s\n", DBASE(i)->id, DBASE(i)->name, ++q);
    }
    for (i = 0; JOBE(i); i++) {
        q = index(JOBE(i)->name, '\0');
        fprintf(stderr, "  %s %s %s\n", JOBE(i)->id, JOBE(i)->name, ++q);
    }
    for (i = 0; STYLS(i); i++) {
        q = index(STYLS(i)->name, '\0');
        fprintf(stderr, "  %s %s %s\n", STYLS(i)->id, STYLS(i)->name, ++q);
    }
    exit(ex);
}

static char *
rediropt(pgoblin_main *options, char *optarg, int redirenv) {
    char *cr;
    if  (redirenv > 0) {
        if  (!(cr = getenv(optarg))) {
            ifBLIN_QV1(options->flags) warnx("No ENV %s", optarg);
            exit(EX_USAGE);
        }
    }else{
        cr = optarg;
    }
    return(cr);
}

static void
cmdin(pgoblin_main *options, int redirenv, pgoblin_prog *pgm, pgoblin_cmd cmd, char *optarg) {
    if  (pgm->curr < pgm->maxx) {
        pgm->code[pgm->curr].cmd = cmd;
        pgm->code[pgm->curr].e.lit = rediropt(options, optarg, redirenv);
        pgm->curr++;
    } else {
        errx(EX_SOFTWARE, "Programm too big");
}   }

static mife_descriptor *
popin(pgoblin_main *options) {
    mife_descriptor *in;
    if  (options->stack[options->sp]) {
        in = mife_open(PGOBLIN_MIFEFLAGS, options->stack[options->sp++]);
    } else {
        in = mife_opef(PGOBLIN_MIFEFLAGS, fileno(stdin));
    }
    return(in);
}

int
main(int argc, char *argv[], char **envp) {
    pgoblin_main *options;
    babolo_opt *opt;
    babolo_parm *bp;
    int c, i, redirenv, ex = EX_OK;
    int dbn = -1, spmax;
    u_int32_t tmp;
    char *o;
    u_char *text;
    size_t txlen;
    mife_descriptor *in;
    pgoblin_prog *pgm;

    if  (!(options = pgoblin_init( BLIN_VER1 | PGOBLIN_DEFAULDB | PGOBLIN_OLDSTYLE
                                 | PGOBLIN_ARGVSTACK | PGOBLIN_ERROR_SENSOR
                                 , fileno(stdout)
        ) )                      ) {
        errx(EX_UNAVAILABLE, "Malloc failed #1");
    }
    if  (!(options->stack = malloc(sizeof(char**) * ((spmax = (argc + 16)) + 1)))) {
        errx(EX_UNAVAILABLE, "Malloc failed #2");
    }
    if (!(pgm = malloc(sizeof(pgoblin_tuple) * PGOBLIN_PROGSIZE + sizeof(pgoblin_prog)))) {
        errx(EX_UNAVAILABLE, "Malloc failed #3");
    }
    options->sp = 0;
    options->stack[0] = NULL;
    options->argv = argv;
    options->envp = envp;
    bzero(pgm, sizeof(pgoblin_tuple) * PGOBLIN_PROGSIZE + sizeof(pgoblin_prog));
    strncpy(pgm->id, "#pGoblin-" VERS, PGOBLIN_STRING_ID_LEN);
    pgm->maxx = PGOBLIN_PROGSIZE;
    pgm->curr = 0;
    pgm->debug = NULL;
    opt = NULL;
    if  (babolo_openopt( Bpars_NONU, &opt, argc, argv
                       , "aAB:c:C:d:D:e:Ef:Fh:INOp:qr:R:s:S:t:U:vVw:zZ?"
        )              ) {
        errx(EX_SOFTWARE, "openopt");
    }
    for (redirenv = 0; (c = babolo_getopt(&opt)) != 0; redirenv--) {
        if  (redirenv < 0) redirenv = 0;
        switch (c) {
        case 'a': options->flags |= PGOBLIN_COPEND_PASS;                                         break;
        case 'A': options->flags |= PGOBLIN_COPEND_APND;                                         break;
        case 'B': if  (dbn >= 0) {
                ;     ifBLIN_QV1(options->flags)
                          warnx("Dublicate DBMS definition %s", babolo_getoptarg(&opt, 0));
                ;     break;
                ; }
                ; for (i = 0; i < pgoblin.dbtablesz; i++) {
                ;     if  (DBASE(i) && DBASE(i)->name) {
                ;         if  (!strcmp(babolo_getoptarg(&opt, 0), DBASE(i)->name)) {
                ;             dbn = i;
                ;             break;
                ; }   }   }
                ; if  (dbn < 0) {
                ;     ifBLIN_QV1(options->flags) warnx("Illegal DBMS %s", babolo_getoptarg(&opt, 0));
                ; }
                ; break;
        case 'c': cmdin(options, redirenv, pgm, PGOBLIN_copyout, babolo_getoptarg(&opt, 0));     break;
        case 'C': cmdin(options, redirenv, pgm, PGOBLIN_copyin, babolo_getoptarg(&opt, 0));      break;
        case 'd': options->conn[0].dbname = rediropt(options, babolo_getoptarg(&opt, 0), redirenv);
                ; break;
        case 'D': cmdin(options, redirenv, pgm, PGOBLIN_perform, babolo_getoptarg(&opt, 0));     break;
        case 'e': ex = pgoblin_parser( options
                                     , (u_char*)rediropt(options, babolo_getoptarg(&opt, 0), redirenv)
                                     , strlen(rediropt(options, babolo_getoptarg(&opt, 0), redirenv))
                                     , &pgm
                                     );
                ; if  (ex) err(EX_DATAERR, "Syntax?");
                ; break;
        case 'E': redirenv = 2;                                                                  break;
        case 'f': options->stack[options->sp] = rediropt(options, babolo_getoptarg(&opt, 0), redirenv);
                ; if  (options->stack[options->sp]) options->stack[++options->sp] = NULL;
                ; break;
        case 'F': if  (options->conn[0].flags & PGOBLIN_INT64TIME) {
                ;     options->conn[0].flags |= PGOBLIN_NEEDINTIME;
                ;     options->conn[0].flags &= ~PGOBLIN_INT64TIME;
                ; } else {
                ;     options->conn[0].flags |= PGOBLIN_FLOATTIME;
                ; }
                ; break;
        case 'h': options->conn[0].host = rediropt(options, babolo_getoptarg(&opt, 0), redirenv);break;
        case 'I': if  (options->conn[0].flags & PGOBLIN_FLOATTIME) {
                ;     options->conn[0].flags |= PGOBLIN_NEEDINTIME;
                ;     options->conn[0].flags &= ~PGOBLIN_FLOATTIME;
                ; } else {
                ;     options->conn[0].flags |= PGOBLIN_INT64TIME;
                ; }
                ; break;
        case 'N': options->flags &= ~PGOBLIN_OLDSTYLE;                                           break;
        case 'O': options->flags |= PGOBLIN_OLDSTYLE;                                            break;
        case 'p': options->conn[0].port = rediropt(options, babolo_getoptarg(&opt, 0), redirenv);break;
        case 'q': options->flags &= ~BLIN_VERMASK;                                               break;
        case 'r': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; bp = babolo_getparm(Bpars_NOEN | Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 'R': if  (!(o = babolo_getoptarg(&opt, 0))) usage(EX_USAGE);
                ; bp = babolo_getparm(Bpars_NOAL | Bpars_NONU, (u_char**)&o, (u_char*)"", 0);
                ; babolo_openopt(0, &opt, bp->argc, (char **)bp->argv, NULL);
                ; break;
        case 's': cmdin(options, redirenv, pgm, PGOBLIN_strselect, babolo_getoptarg(&opt, 0));   break;
        case 'S': cmdin(options, redirenv, pgm, PGOBLIN_binselect, babolo_getoptarg(&opt, 0));   break;
        case 't': options->maxtmout
                = strtoul(rediropt(options, babolo_getoptarg(&opt, 0), redirenv), NULL, 0) * 1000;
                ; break;
        case 'U': options->conn[0].username = rediropt(options, babolo_getoptarg(&opt, 0), redirenv);
                ; break;
        case 'v': BLIN_VERBOSE(options->flags);                                                  break;
        case 'V': if  (options->flags & PGOBLIN_VERIFY) {
                ;     pgoblin_VVdump();
                ; } else {
                ;     options->flags |= PGOBLIN_VERIFY;
                ; }
                ; break;
        case 'w': tmp = strtoul(rediropt(options, babolo_getoptarg(&opt, 0), redirenv), NULL, 0);
                ; options->maxtmout = tmp;
                ; options->flags |= ((tmp > 255) ? 255 : tmp);
                ; break;
        case 'z': /* | PGOBLIN_ERROR_SENSIB not yet */;                                          break;
        case 'Z': /* & ~(PGOBLIN_ERROR_SENSIB | PGOBLIN_ERROR_SENSOR) not yet */;                break;
        case '?': usage(EX_OK);
        default : usage(EX_USAGE);
    }   }

    /** IO_MIFE + **/
    ;;  /** IO_OUT + **/
    ;;  ;;  options->io[0].flags |= options->flags & BLIN_VERMASK;
    ;;  /** IO_OUT - **/
    /** IO_MIFE - **/
    if  (dbn < 0) dbn = PGOBLIN_DB_PGSQL;
    /** R_CONN + **/
    ;;  options->conn[0].flags |= (options->flags & BLIN_VERMASK) | dbn;
    /** R_CONN - **/
    if  (options->flags & PGOBLIN_DEFAULDB) for (i = 1; i < PGOBLIN_REGSIZE; i++) {
        /** R_CONN + **/
        ;;  options->conn[i].flags |= dbn;
        /** R_CONN - **/
    }
    /** R_JOB + **/
    ;;  options->job[0].flags |= options->flags & BLIN_VERMASK;
    /** R_JOB - **/
    /** R_STYLE + **/
    ;;  options->style[0].flags |= options->flags & BLIN_VERMASK;
    /** R_STYLE - **/

    if  (options->flags & PGOBLIN_VERIFY) pgoblin_verify(options);
    options->ap = options->sp;
    while ((o = babolo_getarg(&opt))) {
        options->stack[options->sp++] = o;
        if  (options->sp >= spmax) {
            ifBLIN_QV1(options->flags) warnx("Arg list too grows");
    }   }
    options->stack[options->sp] = NULL;
    options->sp = 0;
    if  (!pgm->curr) {
        if  (!(in = popin(options))) {
            ifBLIN_QV1(options->flags) warn("Programm");
            ex = EX_IOERR;
        } else if ((txlen = mife_read(in, 0, 0)) < 0) {
            ifBLIN_QV1(options->flags) warn("Program read failed");
            ex = EX_IOERR;
        } else if (!(text = mife_get(in, 0))) {
            ifBLIN_QV1(options->flags) warn("Program get failed");
            ex = EX_IOERR;
        } else if (pgoblin_parser(options, text, txlen, &pgm)) {
            ifBLIN_QV1(options->flags) warnx("Parser failed");
            ex = EX_DATAERR;
    }   }
    if  (!ex) {
        /** IO_MIFE + **/
        ;;  if  ((options->io[0].mife = popin(options))) {
        ;;      options->io[0].pid = 0;
        ;;      options->io[0].offset = 0;
        ;;      options->io[0].flags |= PGOBLIN_MIFEDES | PGOBLIN_FREEMIFE;
        ;;  }
        /** IO_MIFE - **/
        if  (options->flags & PGOBLIN_VERIFY) pgoblin_Vdump(options, pgm);
        if  ((ex = pgoblin_execute(options, pgm, NULL, NULL, 0)))
            ifBLIN_QV1(options->flags) warnx("Execute failed");
    }
    exit(ex);
}
